package gregapi.worldgen;

import static gregapi.data.CS.*;
import gregapi.code.ArrayListNoNulls;
import gregapi.code.HashSetNoNulls;
import gregapi.util.WD;

import java.util.List;
import java.util.Random;

import net.minecraft.entity.item.EntityItem;
import net.minecraft.util.AxisAlignedBB;
import net.minecraft.world.World;
import net.minecraft.world.WorldProviderEnd;
import net.minecraft.world.WorldProviderHell;
import net.minecraft.world.biome.BiomeGenBase;
import net.minecraft.world.chunk.Chunk;

/**
 * @author Gregorius Techneticies
 */
public class GT6WorldGenerator {
	public static class WorldGenContainer implements Runnable {
		public final int mMinX, mMinZ, mMaxX, mMaxZ, mDimType;
		public final World mWorld;
		public final Random mRandom;
		public final List<WorldgenObject> mGenNormal;
		public final List<WorldgenOresLarge> mGenLargeOres;
		
		public WorldGenContainer(List aGenNormal, List aGenLargeOres, int aDimType, World aWorld, int aX, int aZ) {
			mMinX = aX; mMinZ = aZ; mMaxX = aX + 15; mMaxZ = aZ + 15;
			mDimType = aDimType;
			mWorld = aWorld;
			mGenNormal = aGenNormal;
			mGenLargeOres = aGenLargeOres;
			mRandom = WD.random(aWorld, aX >> 4, aZ >> 4);
		}
		
		@Override
		public void run() {
			if (!mGenNormal.isEmpty()) {
				Chunk tChunk = mWorld.getChunkFromBlockCoords(mMinX+7, mMinZ+7);
				if (tChunk == null) return;
				BiomeGenBase[][] tBiomes = new BiomeGenBase[16][16];
				HashSetNoNulls<String> tBiomeNames = new HashSetNoNulls<String>();
				for (int i = 0; i < 16; i++) for (int j = 0; j < 16; j++) {
					tBiomes[i][j] = tChunk.getBiomeGenForWorldCoords(i, j, mWorld.provider.worldChunkMgr);
					if (tBiomes[i][j] == null) {
						tBiomes[i][j] = (mDimType == DIM_NETHER ? BiomeGenBase.hell : mDimType == DIM_END ? BiomeGenBase.sky : BiomeGenBase.plains);
					} else {
						tBiomeNames.add(tBiomes[i][j].biomeName);
					}
				}
				
				// Yes, it has to be looped twice in a row, this cannot be optimised into one Loop!
				for (WorldgenObject tWorldGen : mGenNormal) tWorldGen.reset(mWorld, tChunk, mDimType, mMinX, mMinZ, mMaxX, mMaxZ, mRandom, tBiomes, tBiomeNames);
				for (WorldgenObject tWorldGen : mGenNormal) try {if (tWorldGen.enabled(mWorld, mDimType)) tWorldGen.generate(mWorld, tChunk, mDimType, mMinX, mMinZ, mMaxX, mMaxZ, mRandom, tBiomes, tBiomeNames);} catch (Throwable e) {e.printStackTrace(ERR);}
				
				if (mGenLargeOres != null) {
					int tMaxWeight = 0;
					List<WorldgenOresLarge> tList = new ArrayListNoNulls<WorldgenOresLarge>();
					
					for (WorldgenOresLarge tWorldGen : mGenLargeOres) if (tWorldGen.enabled(mWorld, mDimType)) {tMaxWeight += tWorldGen.mWeight; tList.add(tWorldGen);}
					
					if (tMaxWeight > 0 && !tList.isEmpty()) for (int tX=-32; tX<=32; tX+=16) for (int tZ=-32; tZ<=32; tZ+=16) {
						int tChunkX = mMinX+tX, tChunkZ = mMinZ+tZ;
						if (((tChunkX >> 4)+402653184) % 3 == 1 && ((tChunkZ >> 4)+402653184) % 3 == 1) {
							Random aRandom = WD.random(mWorld, tChunkX >> 4, tChunkZ >> 4);
							int tRandomWeight = aRandom.nextInt(tMaxWeight);
							for (WorldgenOresLarge tWorldGen : tList) {
								tRandomWeight -= tWorldGen.mWeight;
								if (tRandomWeight <= 0) {try {tWorldGen.generate(mWorld, tChunk, mMinX, mMinZ, mMaxX, mMaxZ, tChunkX, tChunkZ, mRandom);} catch (Throwable e) {e.printStackTrace(ERR);} break;}
							}
						}
					}
				}
				
				// Kill off every single Item Entity that may have dropped during Worldgen.
				for (EntityItem tEntity : (List<EntityItem>)mWorld.getEntitiesWithinAABB(EntityItem.class, AxisAlignedBB.getBoundingBox(mMinX-32, 0, mMinZ-32, mMinX+48, 256, mMinZ+48))) tEntity.setDead();
				
				tChunk.isModified = T;
			}
		}
	}
	
	private static final List<Runnable> LIST = new ArrayListNoNulls<Runnable>();
	private static boolean LOCK = F;
	public static boolean PFAA = F;
	
	public static void generate(World aWorld, int aX, int aZ, boolean aGalactiCraft) {
		switch(aWorld.provider.dimensionId) {
		case -2147483648	: return;
		case DIM_OVERWORLD	: generate(new WorldGenContainer(PFAA ? GEN_PFAA : GENERATE_STONE ? GEN_GT : GEN_OVERWORLD	, PFAA ? ORE_PFAA : GENERATE_STONE ? null : ORE_OVERWORLD	, DIM_OVERWORLD	, aWorld, aX, aZ)); return;
		case DIM_NETHER		: generate(new WorldGenContainer(GEN_NETHER													, ORE_NETHER												, DIM_NETHER	, aWorld, aX, aZ)); return;
		case DIM_END		: generate(new WorldGenContainer(GEN_END													, ORE_END													, DIM_END		, aWorld, aX, aZ)); return;
		}
		
		if (WD.dimTF(aWorld)) {
			generate(new WorldGenContainer(GEN_TWILIGHT, ORE_TWILIGHT, DIM_TWILIGHT, aWorld, aX, aZ));
			return;
		}
		if (WD.dimERE(aWorld)) {
			generate(new WorldGenContainer(GEN_EREBUS, ORE_EREBUS, DIM_EREBUS, aWorld, aX, aZ));
			return;
		}
		if (WD.dimBTL(aWorld)) {
			generate(new WorldGenContainer(GEN_BETWEENLANDS, ORE_BETWEENLANDS, DIM_BETWEENLANDS, aWorld, aX, aZ));
			return;
		}
		if (WD.dimATUM(aWorld)) {
			generate(new WorldGenContainer(GEN_ATUM, ORE_ATUM, DIM_ATUM, aWorld, aX, aZ));
			return;
		}
		if (WD.dimDD(aWorld)) {
			generate(new WorldGenContainer(GEN_DEEPDARK, ORE_DEEPDARK, DIM_DEEPDARK, aWorld, aX, aZ));
			return;
		}
		
		BiomeGenBase aBiome = aWorld.getBiomeGenForCoords(aX+7, aZ+7);
		if (aBiome == null || BIOMES_VOID.contains(aBiome.biomeName)) return;
		
		if (BIOMES_EREBUS.contains(aBiome.biomeName)) {
			generate(new WorldGenContainer(GEN_EREBUS, ORE_EREBUS, DIM_EREBUS, aWorld, aX, aZ));
			return;
		}
		if (BIOMES_MOON.contains(aBiome.biomeName)) {
			generate(new WorldGenContainer(GEN_MOON, ORE_MOON, DIM_MOON, aWorld, aX, aZ));
			return;
		}
		if (BIOMES_MARS.contains(aBiome.biomeName)) {
			generate(new WorldGenContainer(GEN_MARS, ORE_MARS, DIM_MARS, aWorld, aX, aZ));
			return;
		}
		if (BIOMES_ASTEROIDS.contains(aBiome.biomeName)) {
			generate(new WorldGenContainer(GEN_ASTEROIDS, ORE_ASTEROIDS, DIM_ASTEROIDS, aWorld, aX, aZ));
			return;
		}
		if (aWorld.provider instanceof WorldProviderHell || BIOMES_NETHER.contains(aBiome.biomeName)) {
			generate(new WorldGenContainer(GEN_NETHER, ORE_NETHER, DIM_NETHER, aWorld, aX, aZ));
			return;
		}
		if (aWorld.provider instanceof WorldProviderEnd || BIOMES_END.contains(aBiome.biomeName)) {
			generate(new WorldGenContainer(GEN_END, ORE_END, DIM_END, aWorld, aX, aZ));
			return;
		}
		if (aGalactiCraft || BIOMES_SPACE.contains(aBiome.biomeName)) {
			generate(new WorldGenContainer(GEN_PLANETS, ORE_PLANETS, DIM_PLANETS, aWorld, aX, aZ));
			return;
		}
		generate(new WorldGenContainer(PFAA ? GEN_PFAA : GENERATE_STONE ? GEN_GT : GEN_OVERWORLD, GENERATE_STONE ? null : PFAA ? ORE_PFAA : GENERATE_STONE ? null : ORE_OVERWORLD, DIM_OVERWORLD, aWorld, aX, aZ));
	}
	
	public static void generate(WorldGenContainer aWorldgen) {
		LIST.add(aWorldgen);
		if (!LOCK) {
			LOCK = T;
			while (!LIST.isEmpty()) LIST.remove(LIST.size()-1).run();
			LOCK = F;
		}
	}
}